@@ -15,11 +15,11 @@ class Subsection:
         :param section: a Section instance that contains this Subsection
         """
         self.section = section
+        self.conditional = False
         self.header = content[0].strip().strip('[]').strip()
         if ';' in self.header:
             pos = self.header.index(';')
             self.header = self.header[:pos].strip().strip('[]').strip()
-        self.write_header = self.header if self.header != 'impropers' else 'dihedrals'
         if self.header in Subsection.counter.keys():
             Subsection.counter[self.header] += 1
         else:
@@ -81,15 +81,16 @@ class Subsection:
             return self.entries[n]
         except IndexError:
             raise StopIteration
-    
+
     def add_entry(self, new_entry, position=None):
         """
         Adds a single entry to the subsection, either at the end
         or in a specified position
-        :param new_entry: str, entry to be added
+        :param new_entry: gml.Entry, entry to be added
         :param position: where to add the entry (None is at the end)
         :return: None
         """
+        new_entry.subsection = self
         if position is not None:
             position = int(position)
             self.entries.insert(position, new_entry)
@@ -100,10 +101,12 @@ class Subsection:
         """
         Adds multiple entries to the subsection, either at the end
         or in a specified position
-        :param new_entries_list: list of str, entries to be added
+        :param new_entries_list: list of gml.Entry, entries to be added
         :param position: where to add the entries (None is at the end)
         :return: None
         """
+        for new_entry in new_entries_list:
+            new_entry.subsection = self
         if position is not None:
             position = int(position)
             for new_entry in new_entries_list:
@@ -138,19 +141,23 @@ class SubsectionBonded(Subsection):
     SubsectionBonded contains a subsection with entries corresponding to bonded terms,
     e.g., bonds or dihedrals; should be included in SectionMol
     """
-    n_atoms = {'bonds': 2, 'pairs': 2, 'angles': 3, 'dihedrals': 4, 'impropers': 4,
-               'cmap': 5, 'settles': 1, 'exclusions': 2, 'position_restraints': 1}
+    n_atoms = {'bonds': 2, 'pairs': 2, 'angles': 3, 'dihedrals': 4,
+               'cmap': 5, 'settles': 1, 'exclusions': 2, 'position_restraints': 1,
+               'virtual_sites2': 3, 'constraints': 2}
     
     def __init__(self, content, section):
         super().__init__(content, section)
         self.bkp_entries = None
         self.atoms_per_entry = SubsectionBonded.n_atoms[self.header]
-        self.prmtype = self._check_parm_type()
-        self.label = '{}-{}'.format(self.header, self.prmtype)
+        self.prmtypes = self._check_parm_type()
         self.fstring = "{:5} " * (SubsectionBonded.n_atoms[self.header] + 1) + '\n'
     
     def __repr__(self):
-        return "Subsection {} with interaction type {}".format(self.header, self.prmtype)
+        return "Subsection {} with interaction type {}".format(self.header, ' '.join(self.prmtypes))
+
+    @property
+    def label(self):
+        return '{}-{}'.format(self.header, '_'.join(self.prmtypes))
 
     @property
     def entries_bonded(self):
@@ -195,8 +202,7 @@ class SubsectionBonded(Subsection):
         :param force_all: bool, whether to add params from scratch if some are assigned already
         :return: None
         """
-        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
-                     'impropers': 'dihedraltypes'}
+        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes'}
         subsect_params = [sub for sub in self.section.top.parameters.subsections
                           if sub.header == matchings[self.header]]
         self.bkp_entries = self.entries[:]  # we can't change what we're iterating over, so we modify the copy
@@ -220,8 +226,7 @@ class SubsectionBonded(Subsection):
         :return: list of str, labels of used parameters
         """
         used_parm_entries = []
-        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
-                     'impropers': 'dihedraltypes'}
+        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes', 'cmap': 'cmaptypes'}
         subsect_params = [sub for sub in self.section.top.parameters.subsections if
                           sub.header == matchings[self.header]]
         for entry in self.entries_bonded:
@@ -264,24 +269,28 @@ class SubsectionBonded(Subsection):
                                 pass
         return entries
 
-    def find_missing_ff_params(self, fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False, once=False):
+    def find_missing_ff_params(self, fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False, fix_dummy=False,
+                               once=False):
         """
         Identifies FF parameters that cannot by matched to the existing set of
         bonded interactions, allowing to find & target these for fixing
         :param fix_by_analogy: dict, tells how to choose new parameters from existing equivalent types
         :param fix_B_from_A: bool, chooses missing parameters assuming B-state should have similar types to A-state
         :param fix_A_from_B: bool, chooses missing parameters assuming A-state should have similar types to B-state
+        :param fix_dummy: bool, adds dummy (all-0) parameters
         :param once: bool, only shows the missing parameter once if repeated
         :return: None
         """
-        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
-                     'impropers': 'dihedraltypes'}
+        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes'}
+        # TODO add cmaps?
         subsect_params = [sub for sub in self.section.top.parameters.subsections if
                           sub.header == matchings[self.header]]
         for entry in self.entries_bonded:
-            self._find_missing_ff_params(entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, once)
+            self._find_missing_ff_params(entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, fix_dummy,
+                                         once)
 
-    def _find_missing_ff_params(self, entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, once):
+    def _find_missing_ff_params(self, entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, fix_dummy,
+                                once):
         if (entry.params_state_a and entry.params_state_b) or (entry.params_state_a and not entry.types_state_b):
             return
         int_type = entry.interaction_type
@@ -301,6 +310,9 @@ class SubsectionBonded(Subsection):
                 candid = self._fix_by_analogy(fix_by_analogy, entry.types_state_a, subsect_params, int_type)
                 if candid:
                     entry.params_state_a = candid
+            if fix_dummy:
+                entry.params_state_a = [0.0 for _ in gml.EntryBonded.fstr_suff[(self.header, self.prmtypes[0])]]
+                print("setting dummy parameters: " + str(entry))
         if entry.types_state_b and not entry.params_state_b:
             for subsections in subsect_params:
                 for parm_entry in subsections.entries_param:
@@ -448,13 +460,16 @@ class SubsectionBonded(Subsection):
     
     def _check_parm_type(self):
         """
-        Finds number code for interaction type, e.g. CHARMM uses angletype '5' (urey-bradley)
-        while Amber uses angletype '1' (simple harmonic)
-        :return: str, interaction type
+        Finds number codes for interaction type, e.g. CHARMM uses angletype '5' (urey-bradley)
+        while Amber uses angletype '1' (simple harmonic); this will usually be one number
+        but if subsections are merged, we can have mixed interaction types
+        :return: list of strs, interaction types
         """
+        types = []
         for entry in self.entries_bonded:
-            return entry.interaction_type
-        return '0'
+            if entry.interaction_type not in types:
+                types.append(entry.interaction_type)
+        return ['-1'] if not types else types
 
     def add_type_labels(self):
         for entry in self.entries_bonded:
@@ -481,15 +496,15 @@ class SubsectionParam(Subsection):
     def __init__(self, content, section):
         super().__init__(content, section)
         self.atoms_per_entry = SubsectionParam.n_atoms[self.header]
-        self.prmtype = self._check_parm_type()
-        self.label = '{}-{}'.format(self.header, self.prmtype)
+        self.prmtypes = self._check_parm_type()
+        self.label = '{}-{}'.format(self.header, self.prmtypes)
         self.ordering = {}
         if self.header == 'cmaptypes':
             self._process_cmap()
         
     def __repr__(self):
-        if self.prmtype != '0' or self.header not in ('atomtypes', 'implicit_genborn_params'):
-            return "Subsection {} with interaction type {}".format(self.header, self.prmtype)
+        if self.prmtypes[0] != '-1' or self.header not in ('atomtypes', 'implicit_genborn_params'):
+            return "Subsection {} with interaction type {}".format(self.header, ' '.join(self.prmtypes))
         else:
             return "Subsection {}".format(self.header)
     
@@ -531,7 +546,35 @@ class SubsectionParam(Subsection):
                 used_parm_entries.append(entry.identifier)
         return used_parm_entries
 
-    def _sorting_fn(self, entry):
+    def _combine_entries(self, other_subsection: "gml.SubsectionParam", overwrite: bool = False):
+        """
+        When merging parameter sets, makes sure there are no duplications, and in these cases resolves conflicts
+        :param other_subsection: the gml.Subsection instance from which entries will be added to merge files
+        :param overwrite: bool, whether to accept new parameters in case of a mismatch
+        :return: None
+        """
+        for entry in other_subsection.entries_param:
+            existing = [e for e in self.entries_param if e.types == entry.types]
+            if len(existing) == 1:
+                ex_entry = existing[0]
+                if ex_entry.params == entry.params and ex_entry.interaction_type == entry.interaction_type \
+                        and ex_entry.modifiers == entry.modifiers:
+                    continue
+                else:
+                    if overwrite:
+                        ex_entry.params = entry.params
+                        ex_entry.interaction_type = entry.interaction_type
+                        ex_entry.modifiers = entry.modifiers
+                    else:
+                        print(f"Found mismatch between original entry {str(ex_entry)} and new {entry}, keeping the "
+                              f"original; if this is not desired, set overwrite=True")
+            elif len(existing) == 0:
+                self.add_entry(entry)
+            else:
+                raise RuntimeError(f"Found multiple entries matching the types of {entry.types} in section"
+                                   f"{self.header}, make sure this is intentional")
+
+    def _sorting_fn(self, entry: "gml.Entry") -> int:
         """
         Comments should go first, then we sort based on first, second,
         ... column of the section
@@ -551,15 +594,11 @@ class SubsectionParam(Subsection):
         :return: str, interaction type
         """
         if self.header not in SubsectionParam.n_atoms.keys() or self.header in ['atomtypes', 'implicit_genborn_params']:
-            return '0'
+            return '-1'
         npar = SubsectionParam.n_atoms[self.header]
-        for entry in self:
-            try:
-                return entry.interaction_type
-            except:
-                if len(entry.content) > npar and isinstance(entry, gml.EntryParam):
-                    return entry.content[npar]
-        return '0'
+        types = list({entry.content[npar] if len(entry.content) > npar else entry.interaction_type
+                      for entry in self.entries_param})
+        return '-1' if not types else types
 
     def get_opt_dih(self, types=False):
         dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
@@ -646,9 +685,22 @@ class SubsectionAtom(Subsection):
         so only fill in the values when explicitly asked to
         :return: None
         """
+        # TODO should they be @properties, called only when needed? EITHER WAY, REVISE
         if not self.name_to_num or force_update:
             self.name_to_num, self.num_to_name, self.num_to_type, self.num_to_type_b = self._mol_type_nums()
 
+    def check_defined_types(self):
+        defined_types = self.section.top.parameters.get_subsection('atomtypes')
+        typelist = {ent.types[0] for ent in defined_types.entries_param}
+        for atom in self.entries_atom:
+            if atom.type not in typelist:
+                print(f'Couldn\'t find definition of atom type {atom.type} (atom {atom.num} in molecule '
+                      f'{self.section.mol_name}) in parameters')
+            if atom.type_b:
+                if atom.type_b not in typelist:
+                    print(f'Couldn\'t find definition of atom type {atom.type_b} (atom {atom.num} in molecule '
+                          f'{self.section.mol_name}) in parameters')
+
     def _mol_type_nums(self):
         """
         Provides bindings between atomnumber and atomtype
