@@ -5,5 +5,11 @@ from gromologist.Entries import Entry, EntryAtom, EntryBonded, EntryParam
 from gromologist.Pdb import Pdb, Atom, Traj
 from gromologist.DihOpt import DihOpt
 from gromologist.Gmx import *
+from gromologist.Utils import *
 from gromologist.Parser import SelectionParser
-from gromologist.Mutant import ProteinMutant
\ No newline at end of file
+from gromologist.Mutant import ProteinMutant
+from gromologist.ThDiff import *
+try:
+    from gromologist.Crooks import *
+except ModuleNotFoundError:
+    pass
